@@ -2,12 +2,13 @@ from __future__ import annotations
 
 from dataclasses import dataclass
 from dataclasses import field as dc_field
+from dataclasses import replace
 from itertools import zip_longest
 from math import ceil, log2
-from typing import Any, Callable, overload
+from typing import Any, Callable, TypeVar, overload
 
 from vskernels import Catrom, Kernel, KernelT, Scaler, ScalerT
-from vstools import core, inject_self, vs
+from vstools import core, inject_self, vs, vs_object
 
 from .enums import AADirection
 
@@ -18,12 +19,9 @@ __all__ = [
 ]
 
 
-class _SingleInterpolate:
+class _SingleInterpolate(vs_object):
     _shift: float
 
-    def __post_init__(self) -> None:
-        ...
-
     def interpolate(self, clip: vs.VideoNode, double_y: bool, **kwargs: Any) -> vs.VideoNode:
         raise NotImplementedError
 
@@ -61,6 +59,12 @@ class _Antialiaser(_SingleInterpolate):
 
         return self._shifter.scale(inter, clip.width, clip.height, shift)
 
+    def copy(self: SelfAntialiaser, **kwargs: Any) -> SelfAntialiaser:
+        return replace(self, **kwargs)
+
+
+SelfAntialiaser = TypeVar('SelfAntialiaser', bound=_Antialiaser)
+
 
 class _FullInterpolate(_SingleInterpolate):
     def is_full_interpolate_enabled(self, x: bool, y: bool) -> bool:
