@@ -2,23 +2,24 @@ from __future__ import annotations
 
 from math import ceil
 from pathlib import Path
-from typing import TYPE_CHECKING
+from typing import TYPE_CHECKING, Any
 
 from vsexprtools import norm_expr_planes
 from vskernels import Catrom, Scaler, ScalerT, Spline144
-from vsmask.edge import EdgeDetect, Prewitt, ScharrTCanny
-from vsrgtools import RepairMode, box_blur, contrasharpening_median, median_clips, repair
+from vsmasktools import EdgeDetect, EdgeDetectT, Prewitt, ScharrTCanny
+from vsrgtools import RepairMode, box_blur, contrasharpening_median, median_clips, repair, unsharp_masked
 from vstools import (
-    MISSING, CustomOverflowError, CustomRuntimeError, PlanesT, check_variable, core, get_depth, get_peak_value, get_w,
-    get_y, join, normalize_planes, scale_8bit, scale_value, split, vs
+    MISSING, CustomOverflowError, CustomRuntimeError, MissingT, PlanesT, check_ref_clip, check_variable, core,
+    get_depth, get_peak_value, get_w, get_y, join, normalize_planes, scale_8bit, scale_value, split, vs
 )
 
 from .abstract import Antialiaser, SingleRater
-from .antialiasers import Eedi3, Eedi3SR, Nnedi3SR, Nnedi3SS
+from .antialiasers import Eedi3, Eedi3SR, Nnedi3SR, Nnedi3SS, Znedi3
 from .enums import AADirection
 from .mask import resize_aa_mask
 
 __all__ = [
+    'pre_aa',
     'upscaled_sraa',
     'transpose_aa',
     'clamp_aa', 'masked_clamp_aa',
@@ -27,6 +28,25 @@ __all__ = [
 ]
 
 
+def pre_aa(
+    clip: vs.VideoNode, radius: int = 1, strength: int = 100,
+    aa: type[Antialiaser] | Antialiaser = Znedi3(0, pscrn=1),
+    **kwargs: Any
+) -> vs.VideoNode:
+    if isinstance(aa, Antialiaser):
+        aa = aa.copy(field=3, **kwargs)
+    else:
+        aa = aa(field=3, **kwargs)
+
+    for _ in range(2):
+        bob = aa.interpolate(clip, False)
+        sharp = unsharp_masked(clip, radius, strength)
+        limit = median_clips(sharp, clip, bob[::2], bob[1::2])
+        clip = limit.std.Transpose()
+
+    return clip
+
+
 def upscaled_sraa(
     clip: vs.VideoNode, rfactor: float = 1.5,
     width: int | None = None, height: int | None = None,
@@ -157,7 +177,7 @@ def clamp_aa(
 
 def masked_clamp_aa(
     clip: vs.VideoNode, strength: float = 1,
-    mthr: float = 0.25, mask: vs.VideoNode | EdgeDetect | None = None,
+    mthr: float = 0.25, mask: vs.VideoNode | EdgeDetectT | None = None,
     weak_aa: SingleRater | None = None, strong_aa: SingleRater | None = None,
     opencl: bool | None = True
 ) -> vs.VideoNode:
@@ -178,13 +198,10 @@ def masked_clamp_aa(
 
     work_clip, *chroma = split(clip)
 
-    if mask is None:
-        mask = ScharrTCanny()
-
-    if isinstance(mask, EdgeDetect):
+    if not isinstance(mask, vs.VideoNode):
         bin_thr = scale_value(mthr, 32, get_depth(clip))
 
-        mask = mask.edgemask(work_clip)
+        mask = ScharrTCanny.ensure_obj(mask).edgemask(work_clip)  # type: ignore
         mask = mask.std.Binarize(bin_thr)
         mask = mask.std.Maximum()
         mask = box_blur(mask)
@@ -207,7 +224,7 @@ def masked_clamp_aa(
 
     clamped = clamp_aa(work_clip, weak, strong, strength)
 
-    merged = work_clip.std.MaskedMerge(clamped, mask)
+    merged = work_clip.std.MaskedMerge(clamped, mask)  # type: ignore
 
     if not chroma:
         return merged
@@ -252,25 +269,26 @@ def fine_aa(
 
 
 if TYPE_CHECKING:
+    from vsdenoise import Prefilter
     from vsscale import SSIM, FSRCNNXShader, FSRCNNXShaderT, ShaderFile
 
     def based_aa(
         clip: vs.VideoNode, rfactor: float = 2.0,
-        mask_thr: float = 60, lmask: vs.VideoNode | EdgeDetect = Prewitt(),
+        mask_thr: int = 60, lmask: vs.VideoNode | EdgeDetectT = Prewitt,
         downscaler: ScalerT = SSIM,
         supersampler: ScalerT | FSRCNNXShaderT | ShaderFile | Path = FSRCNNXShader.x56,
         antialiaser: Antialiaser = Eedi3(0.125, 0.25, vthresh0=12, vthresh1=24, field=1, sclip_aa=None),
-        show_mask: bool = False
+        prefilter: Prefilter | vs.VideoNode = Prefilter.NONE, show_mask: bool = False, **kwargs: Any
     ) -> vs.VideoNode:
         ...
 else:
     def based_aa(
         clip: vs.VideoNode, rfactor: float = 2.0,
-        mask_thr: float = 60, lmask: vs.VideoNode | EdgeDetect = Prewitt(),
-        downscaler: ScalerT | MISSING = MISSING,
-        supersampler: ScalerT | FSRCNNXShaderT | ShaderFile | Path | MISSING = MISSING,
+        mask_thr: int = 60, lmask: vs.VideoNode | EdgeDetectT = Prewitt,
+        downscaler: ScalerT | MissingT = MISSING,
+        supersampler: ScalerT | FSRCNNXShaderT | ShaderFile | Path | MissingT = MISSING,
         antialiaser: Antialiaser = Eedi3(0.125, 0.25, vthresh0=12, vthresh1=24, field=1, sclip_aa=None),
-        show_mask: bool = False
+        prefilter: Prefilter | vs.VideoNode | MissingT = MISSING, show_mask: bool = False, **kwargs: Any
     ) -> vs.VideoNode:
         try:
             from vsscale import SSIM, FSRCNNXShader, PlaceboShader  # noqa: F811
@@ -279,33 +297,48 @@ else:
                 'You\'re missing the "vsscale" package! You can install it with "pip install vsscale".', based_aa
             )
 
+        try:
+            from vsdenoise import Prefilter  # noqa: F811
+        except ModuleNotFoundError:
+            raise CustomRuntimeError(
+                'You\'re missing the "vsdenoise" package! You can install it with "pip install vsdenoise".', based_aa
+            )
+
         if downscaler is MISSING:
             downscaler = SSIM
 
         if supersampler is MISSING:
             supersampler = FSRCNNXShader.x56
 
+        if prefilter is MISSING:
+            prefilter = Prefilter.NONE
+
         assert check_variable(clip, based_aa)
 
         aaw = (round(clip.width * rfactor) + 1) & ~1
         aah = (round(clip.height * rfactor) + 1) & ~1
 
         clip_y = get_y(clip)
+        if isinstance(prefilter, vs.VideoNode):
+            wclip_y = get_y(prefilter)
+            check_ref_clip(clip_y, wclip_y, based_aa)
+        else:
+            wclip_y = prefilter(clip_y, **kwargs)
 
         if rfactor < 1.0:
             raise CustomOverflowError(
                 f'"rfactor" must be bigger than 1.0! ({rfactor})', based_aa
             )
 
-        if isinstance(lmask, EdgeDetect):
+        if not isinstance(lmask, vs.VideoNode):
+            lmask = EdgeDetect.ensure_obj(lmask, based_aa)
+
             if mask_thr > 255:
                 raise CustomOverflowError(
                     f'"mask_thr" must be less or equal than 255! ({mask_thr})', based_aa
                 )
 
-            mask_thr = scale_8bit(clip, mask_thr)
-
-            lmask = lmask.edgemask(clip_y).std.Binarize(mask_thr)
+            lmask = lmask.edgemask(wclip_y).std.Binarize(scale_8bit(wclip_y, mask_thr))
             lmask = box_blur(lmask.std.Maximum()).std.Limiter()
 
         if show_mask:
@@ -314,22 +347,30 @@ else:
         if isinstance(supersampler, (str, Path)):
             supersampler = PlaceboShader(supersampler)
 
+        supersampler = Scaler.ensure_obj(supersampler, based_aa)
+        downscaler = Scaler.ensure_obj(downscaler, based_aa)
+
         mclip_up = resize_aa_mask(lmask, aaw, aah)
 
         aa = clip_y.std.Transpose()
         aa = supersampler.scale(aa, aa.width * ceil(rfactor), aa.height * ceil(rfactor))
-        aa = downscaler.scale(aa, aah, aaw)
+        aa = waa = downscaler.scale(aa, aah, aaw)
 
-        def _aa_mclip(clip: vs.VideoNode, mclip: vs.VideoNode) -> vs.VideoNode:
+        if clip_y is not wclip_y:
+            waa = wclip_y.std.Transpose()
+            waa = supersampler.scale(waa, waa.width * ceil(rfactor), waa.height * ceil(rfactor))
+            waa = downscaler.scale(waa, aah, aaw)
+
+        def _aa_mclip(clip: vs.VideoNode, wclip: vs.VideoNode, mclip: vs.VideoNode) -> vs.VideoNode:
             return core.std.Expr([
-                clip, antialiaser.interpolate(clip, False, sclip=clip), mclip
+                clip, antialiaser.interpolate(wclip, False, sclip=wclip), mclip
             ], 'z y x ?')
 
-        aa = _aa_mclip(aa, mclip_up.std.Transpose()).std.Transpose()
+        aa = _aa_mclip(aa, waa, mclip_up.std.Transpose())
 
-        aa = _aa_mclip(aa, mclip_up)
+        aa = _aa_mclip(aa.std.Transpose(), waa.std.Transpose(), mclip_up)
 
-        aa = downscaler.scale(aa, clip.width, clip.height)
+        aa = downscaler.scale(aa, clip_y.width, clip_y.height)
 
         aa_merge = clip_y.std.MaskedMerge(aa, lmask)
 
